@@ -32,6 +32,8 @@
 #include <glibtop/mem.h>
 #include <glibtop/swap.h>
 #include <glibtop/procmap.h>
+#define WNCK_I_KNOW_THIS_IS_UNSTABLE
+#include <libwnck/libwnck.h>
 #include <sys/stat.h>
 #include <pwd.h>
 #include "procman.h"
@@ -99,6 +101,10 @@ sort_ints (GtkTreeModel *model, GtkTreeIter *itera, GtkTreeIter *iterb, gpointer
 		a = infoa->memrss;
 		b = infob->memrss;
 		break;
+        case COL_MEMXSERVER:
+		a = infoa->memxserver;
+		b = infob->memxserver;
+		break;
 	default:
 		return 0;
 		break;
@@ -126,6 +132,7 @@ proctable_new (ProcData *data)
 				 N_("User"), N_("Status"),
 				 N_("Memory"), N_("VM Size"), N_("Resident Memory"),
 				 N_("Shared Memory"), N_("RSS Memory"),
+                                 N_("X Server Memory"),
 				 /* xgettext:no-c-format */
 				 N_("% CPU"),
 				 N_("Nice"), N_("ID"), NULL, "POINTER"};
@@ -139,7 +146,7 @@ proctable_new (ProcData *data)
 	model = gtk_tree_store_new (NUM_COLUMNS,  G_TYPE_STRING, 
 				    G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING,
 				    G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING,
-				    G_TYPE_STRING, G_TYPE_STRING,
+				    G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING,
 				    G_TYPE_INT, G_TYPE_INT, G_TYPE_INT,
 				    GDK_TYPE_PIXBUF, G_TYPE_POINTER);		    
   	
@@ -166,17 +173,21 @@ proctable_new (ProcData *data)
 	gtk_tree_view_column_set_title (column, _(title[0]));
 	gtk_tree_view_column_set_sort_column_id (column, COL_NAME);
 	gtk_tree_view_column_set_resizable (column, TRUE);
+	gtk_tree_view_column_set_min_width (column, 1);
 	gtk_tree_view_append_column (GTK_TREE_VIEW (proctree), column);
 	gtk_tree_view_set_expander_column (GTK_TREE_VIEW (proctree), column);
   	
   	for (i = 1; i < NUM_COLUMNS - 2; i++) {
   		cell_renderer = gtk_cell_renderer_text_new ();
-  		column = gtk_tree_view_column_new_with_attributes (_(title[i]),
-						    		   cell_renderer,
-						     		   "text", i,
-						     		   NULL);
-		gtk_tree_view_column_set_sort_column_id (column, i);
+  		column = gtk_tree_view_column_new ();
+  		gtk_tree_view_column_pack_start (column, cell_renderer, FALSE);
+  		gtk_tree_view_column_set_attributes (column, cell_renderer,
+					     	"text", i,
+					     	NULL);
+		gtk_tree_view_column_set_title (column, _(title[i]));
+  		gtk_tree_view_column_set_sort_column_id (column, i);
 		gtk_tree_view_column_set_resizable (column, TRUE);
+		gtk_tree_view_column_set_min_width (column, 1);
 		gtk_tree_view_append_column (GTK_TREE_VIEW (proctree), column);
 	}
 	
@@ -222,6 +233,11 @@ proctable_new (ProcData *data)
 					 sort_ints,
 					 GINT_TO_POINTER (COL_MEMRSS),
 					 NULL);
+        gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (model),
+					 COL_MEMXSERVER,
+					 sort_ints,
+					 GINT_TO_POINTER (COL_MEMXSERVER),
+					 NULL);
 	procdata->tree = proctree;
 	
 	procman_get_tree_state (procdata->client, proctree, "/apps/procman/proctree");
@@ -229,17 +245,16 @@ proctable_new (ProcData *data)
 	g_signal_connect (G_OBJECT (gtk_tree_view_get_selection (GTK_TREE_VIEW (proctree))), 
 			  "changed",
 			  G_CALLBACK (cb_row_selected), procdata);
+	g_signal_connect (G_OBJECT (proctree), "row_activated",
+			  G_CALLBACK (cb_tree_row_activated), procdata);
+	g_signal_connect (G_OBJECT (proctree), "popup_menu",
+			  G_CALLBACK (cb_tree_popup_menu), procdata);
 	g_signal_connect (G_OBJECT (proctree), "button_press_event",
 			  G_CALLBACK (cb_tree_button_pressed), procdata);
-	g_signal_connect (G_OBJECT (proctree), "key_press_event",
-			  G_CALLBACK (cb_tree_key_press), procdata);
 			  
 	return scrolled;
-
 }
 
-
-
 static void
 proctable_free_info (ProcInfo *info)
 {
@@ -256,8 +271,6 @@ proctable_free_info (ProcInfo *info)
 	g_free (info);
 }
 
-
-
 static void
 get_process_status (ProcInfo *info, char *state)
 {
@@ -326,6 +339,29 @@ get_process_name (ProcData *procdata, ProcInfo *info, gchar *cmd, gchar *args)
 
 }
 
+ProcInfo *
+proctable_find_process (gint pid, gchar *name, ProcData *procdata)
+{
+
+	GList *list = procdata->info;
+	
+	while (list) {
+		ProcInfo *info = list->data;
+		
+		if (pid == info->pid)
+			return info;
+		
+		if (name) {	
+			if (g_strcasecmp (name, info->name) == 0)
+				return info;
+		}
+		
+		list = g_list_next (list);
+	}
+
+	return NULL;
+}
+
 static ProcInfo *
 find_parent (ProcData *data, gint pid)
 {
@@ -348,51 +384,18 @@ find_parent (ProcData *data, gint pid)
 	return NULL;
 }
 
-
-/* He he. Check to see if the process links to libX11. */
-static gboolean
-is_graphical (ProcInfo *info)
-{
-	glibtop_map_entry *memmaps;
-	glibtop_proc_map procmap;
-	char *string=NULL;
-	gint i;
-	
-	memmaps = glibtop_get_proc_map (&procmap, info->pid);
-	
-	for (i = 0; i < procmap.number; i++) {
-		if (memmaps [i].flags & (1 << GLIBTOP_MAP_ENTRY_FILENAME)) {
-			string = strstr (memmaps[i].filename, "libX11");
-			if (string) {
-				g_free (memmaps);
-				return TRUE;
-			}
-		}
-	}
-	
-	g_free (memmaps);
-	
-	return FALSE;
-}
-
-
 void
 insert_info_to_tree (ProcInfo *info, ProcData *procdata)
 {
 	GtkTreeModel *model;
 	GtkTreeIter row;
 	gchar *name;
-	gchar *mem, *vmsize, *memres, *memshared, *memrss;
+	gchar *mem, *vmsize, *memres, *memshared, *memrss, *memxserver;
 	
 	/* Don't show process if it is not running */
-	if (procdata->config.whose_process == RUNNING_PROCESSES && 
+	if ((procdata->config.whose_process == ACTIVE_PROCESSES) && 
 	    (!info->running))
 		return;
-		
-	/* crazy hack to see if process links to libX11 */	
-	/*if (!is_graphical (info))
-		return;*/
-
 
 	/* Don't show processes that user has blacklisted */
 	if (is_process_blacklisted (procdata, info->name))
@@ -411,10 +414,10 @@ insert_info_to_tree (ProcInfo *info, ProcData *procdata)
 		name = g_strdup (info->name);
 			
 	model = gtk_tree_view_get_model (GTK_TREE_VIEW (procdata->tree));
-	if (info->has_parent && procdata->config.show_tree) {
-		GtkTreePath *parent_node = gtk_tree_model_get_path (model, &info->parent_node);
+	if (info->parent && procdata->config.show_tree && info->parent->visible) {
+		GtkTreePath *parent_node = gtk_tree_model_get_path (model, &info->parent->node);
 		
-		gtk_tree_store_insert (GTK_TREE_STORE (model), &row, &info->parent_node, 0);
+		gtk_tree_store_insert (GTK_TREE_STORE (model), &row, &info->parent->node, 0);
 		if (!gtk_tree_view_row_expanded (GTK_TREE_VIEW (procdata->tree), parent_node))     
 			gtk_tree_view_expand_row (GTK_TREE_VIEW (procdata->tree),
 					  parent_node,
@@ -430,6 +433,7 @@ insert_info_to_tree (ProcInfo *info, ProcData *procdata)
 	memres = get_size_string (info->memres);
 	memshared = get_size_string (info->memshared);
 	memrss = get_size_string (info->memrss);
+        memxserver = get_size_string (info->memxserver);
 
         /* COL_POINTER must be set first, because GtkTreeStore
          * will call sort_ints as soon as we set the column
@@ -448,6 +452,7 @@ insert_info_to_tree (ProcInfo *info, ProcData *procdata)
                             COL_MEMRES, memres,
                             COL_MEMSHARED, memshared,
                             COL_MEMRSS, memrss,
+                            COL_MEMXSERVER, memxserver,
                             COL_CPU, info->cpu,
                             COL_PID, info->pid,
                             COL_NICE, info->nice,
@@ -457,16 +462,15 @@ insert_info_to_tree (ProcInfo *info, ProcData *procdata)
 	g_free (memres);
 	g_free (memshared);
 	g_free (memrss);
+        g_free (memxserver);
 	g_free (name);
 	
 	info->node = row;
 	info->visible = TRUE;
 }
 
-
-
-/* Kind of a hack. When a parent process is removed we remove all the info
-** pertaining to the child processes and then readd them later
+/* Removing a node with children - make sure the children are queued
+** to be readded.
 */
 static void
 remove_children_from_tree (ProcData *procdata, GtkTreeModel *model,
@@ -484,11 +488,10 @@ remove_children_from_tree (ProcData *procdata, GtkTreeModel *model,
 		if (child_info) {
 			if (procdata->selected_process == child_info) 
 				procdata->selected_process = NULL;
-			procdata->info = g_list_remove (procdata->info, child_info);
-			proctable_free_info (child_info);
+			child_info ->queue = NEEDS_ADDITION;
+			child_info->visible = FALSE;
 		}
 	} while (gtk_tree_model_iter_next (model, parent));
-		
 }
 
 void
@@ -497,6 +500,7 @@ remove_info_from_tree (ProcInfo *info, ProcData *procdata)
 	GtkTreeModel *model;
 	GtkTreeIter iter, child;
 	GtkTreePath *node;
+	GList *children;
 	
 	g_return_if_fail (info);
 	
@@ -505,18 +509,39 @@ remove_info_from_tree (ProcInfo *info, ProcData *procdata)
 	
 	model = gtk_tree_view_get_model (GTK_TREE_VIEW (procdata->tree));
 	iter = info->node;
-	
-	/* remove all children from tree. They will be added again in refresh_list */
-	if (gtk_tree_model_iter_children (model, &child, &iter))
-		remove_children_from_tree (procdata, model, &child);
-	
+
 	if (procdata->selected_process == info) 
 		procdata->selected_process = NULL;
 	
 	gtk_tree_store_remove (GTK_TREE_STORE (model), &iter);
 	
 	info->visible = FALSE;	
+}
+
+static void
+remove_info_from_list (ProcInfo *info, ProcData *procdata)
+{
+	GList *children = info->children;
 		
+	/* Remove info from parent list */
+	if (info->parent)
+		info->parent->children = g_list_remove (info->parent->children, info);
+
+	/* Add any children to parent list */	
+	while (children) {
+		ProcInfo *child_info = children->data;
+		
+		if (info->parent)	
+			info->parent->children = g_list_append (info->parent->children, child_info);
+		child_info->parent = info->parent;
+			
+		children = g_list_next (children);
+	}
+	
+	/* Remove from list */
+	procdata->info = g_list_remove (procdata->info, info);
+	proctable_free_info (info);
+	
 }
 
 static void
@@ -528,10 +553,11 @@ update_info (ProcData *procdata, ProcInfo *info, gint pid)
 	glibtop_proc_mem procmem;
 	glibtop_proc_uid procuid;
 	glibtop_proc_time proctime;
-	gchar *mem, *vmsize, *memres, *memshared, *memrss;
+	gchar *mem, *vmsize, *memres, *memshared, *memrss, *memxserver;
 	gint newcputime;
 	gint pcpu;
-	
+        WnckResourceUsage xresources;
+        
 	glibtop_get_proc_state (&procstate, pid);
 	glibtop_get_proc_mem (&procmem, pid);
 	glibtop_get_proc_uid (&procuid, pid);
@@ -539,11 +565,22 @@ update_info (ProcData *procdata, ProcInfo *info, gint pid)
 	newcputime = proctime.utime + proctime.stime;
 	model = gtk_tree_view_get_model (GTK_TREE_VIEW (procdata->tree));
 
+        wnck_pid_read_resource_usage (gdk_screen_get_display (gdk_screen_get_default ()),
+                                      pid,
+                                      &xresources);
+
 	info->mem = procmem.size;
 	info->vmsize = procmem.vsize;
 	info->memres = procmem.resident;
 	info->memshared = procmem.share;
 	info->memrss = procmem.rss;
+        info->memxserver = xresources.total_bytes_estimate;
+
+        /* Possible enhancement - also SUBTRACT this value
+         * from the memory of the X server itself.
+         */
+        info->mem += info->memxserver;
+        
 	if (total_time != 0)
 		pcpu = ( newcputime - info->cpu_time_last ) * 100 / total_time;
 	else 
@@ -551,11 +588,11 @@ update_info (ProcData *procdata, ProcInfo *info, gint pid)
 	info->cpu_time_last = newcputime;
 	info->cpu = pcpu;	
 	info->nice = procuid.nice;
-		
-	get_process_status (info, &procstate.state);
 
-	if (procdata->config.whose_process == RUNNING_PROCESSES)
-	{
+	get_process_status (info, (char *) &procstate.state);
+
+	if (procdata->config.whose_process == ACTIVE_PROCESSES)	{
+
 		/* process started running */
 		if (info->running && (!info->visible)) {
 			insert_info_to_tree (info, procdata);
@@ -586,6 +623,7 @@ update_info (ProcData *procdata, ProcInfo *info, gint pid)
 		memres = get_size_string (info->memres);
 		memshared = get_size_string (info->memshared);
 		memrss = get_size_string (info->memrss);
+                memxserver = get_size_string (info->memxserver);
 		
 		gtk_tree_store_set (GTK_TREE_STORE (model), &info->node, 
 				    COL_STATUS, info->status,
@@ -594,6 +632,7 @@ update_info (ProcData *procdata, ProcInfo *info, gint pid)
 				    COL_MEMRES, memres,
 			            COL_MEMSHARED, memshared,
 				    COL_MEMRSS, memrss,
+                                    COL_MEMXSERVER, memxserver,
 				    COL_CPU, info->cpu,
 				    COL_NICE, info->nice,
 			 	   -1);
@@ -602,8 +641,8 @@ update_info (ProcData *procdata, ProcInfo *info, gint pid)
 		g_free (memres);
 		g_free (memshared);
 		g_free (memrss);
+                g_free (memxserver);
 	}
-		
 }
 
 static ProcInfo *
@@ -619,7 +658,7 @@ get_info (ProcData *procdata, gint pid)
 	struct passwd *pwd;
 	gchar *arguments;
 	gint newcputime, i;
-	
+        WnckResourceUsage xresources;
 	
 	glibtop_get_proc_state (&procstate, pid);
 	pwd = getpwuid (procstate.uid);
@@ -629,6 +668,10 @@ get_info (ProcData *procdata, gint pid)
 	glibtop_get_proc_time (&proctime, pid);
 	newcputime = proctime.utime + proctime.stime;
 
+        wnck_pid_read_resource_usage (gdk_screen_get_display (gdk_screen_get_default ()),
+                                      pid,
+                                      &xresources);
+        
 	arguments = glibtop_get_proc_args (&procargs, pid, 0);	
 	get_process_name (procdata, info, procstate.cmd, arguments);
 	if (arguments)
@@ -653,16 +696,27 @@ get_info (ProcData *procdata, gint pid)
 	info->memres = procmem.resident;
 	info->memshared = procmem.share;
 	info->memrss = procmem.rss;
+        info->memxserver = xresources.total_bytes_estimate;
+
+        /* Possible enhancement - also SUBTRACT this value
+         * from the memory of the X server itself.
+         */
+        info->mem += info->memxserver;
+        
 	info->cpu = 0;
 	info->pid = pid;
 	info->parent_pid = procuid.ppid;
 	info->cpu_time_last = newcputime;
 	info->nice = procuid.nice;
-	get_process_status (info, &procstate.state);
+	get_process_status (info, (char *) &procstate.state);
 	
 	info->pixbuf = pretty_table_get_icon (procdata->pretty_table, info->name, pid);
 	
-	parentinfo = find_parent (procdata, info->parent_pid);
+	info->parent = find_parent (procdata, info->parent_pid);
+	if (info->parent)
+		info->parent->children = g_list_append (info->parent->children, info);
+
+#if 0
 	if (parentinfo) {
 		/* Ha Ha - don't expand different threads - check to see if parent has
 		** same name and same mem usage - I don't know if this is too smart though.
@@ -688,104 +742,95 @@ get_info (ProcData *procdata, gint pid)
 		info->has_parent = FALSE;
 		info->is_thread = FALSE;
 	}
+#endif
 	
 	info->visible = FALSE;
 	
 	return info;
 }
 
-ProcInfo *
-proctable_find_process (gint pid, gchar *name, ProcData *procdata)
+static gboolean
+find_match_in_new_list (gint pid, unsigned *pid_list, gint n)
 {
-
-	GList *list = procdata->info;
+	gint i = 0;
 	
-	while (list) {
-		ProcInfo *info = list->data;
-		
-		if (pid == info->pid)
-			return info;
-		
-		if (name) {	
-			if (g_strcasecmp (name, info->name) == 0)
-				return info;
-		}
-		
-		list = g_list_next (list);
+	for (i=0; i<n; i++){
+		if (pid_list[i] == pid) return TRUE;	
 	}
 
-	return NULL;
+	return FALSE;
 }
 
 static void
 refresh_list (ProcData *data, unsigned *pid_list, gint n)
 {
 	ProcData *procdata = data;
-	GList *list = procdata->info;
+	GList *list, *removal_list = NULL;
+	GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW (procdata->tree));
 	gint i = 0;
 	
-	while (i < n)
-	{
-		ProcInfo *oldinfo;
+	/* Add or update processes */
+	while (i < n) {
+		ProcInfo *info;
 		
-		/* New process with higher pid than any previous one */
-		if (!list)
-		{
-			ProcInfo *info;
-			
+		info = proctable_find_process (pid_list[i], NULL, procdata);
+		if (!info) {
 			info = get_info (procdata, pid_list[i]);
-			insert_info_to_tree (info, procdata);
+			info->queue = NEEDS_ADDITION;
 			procdata->info = g_list_append (procdata->info, info);
-			
-			i++;
-			continue;
 		}
-		else
-			oldinfo = list->data;
-		/* new process */	
-		if (pid_list[i] < oldinfo->pid)
-		{
-			ProcInfo *info;
-			
-			info = get_info (procdata, pid_list[i]);
-			insert_info_to_tree (info, procdata);
-			procdata->info = g_list_insert (procdata->info, info, i);
-			
-			i++;
-		}
-		/* existing process */
-		else if (pid_list[i] == oldinfo->pid)
-		{
-			update_info (procdata, oldinfo, oldinfo->pid);
-			
-			list = g_list_next (list);
-			i++;
-		}
-		/* process no longer exists */
-		else if (pid_list[i] > oldinfo->pid)
-		{		
-			remove_info_from_tree (oldinfo, procdata);
-			list = g_list_next (list);
-			procdata->info = g_list_remove (procdata->info, oldinfo);
-			proctable_free_info (oldinfo);
-											
+		else {
+			update_info (procdata, info, info->pid);
 		}
+		
+		i++;
 	}
 	
-	
-	/* Get rid of any tasks at end of list that have ended */
-	while (list)
-	{
-		ProcInfo *oldinfo;
-		oldinfo = list->data;
-		remove_info_from_tree (oldinfo, procdata);
+	/* Remove processes */
+	list = procdata->info;
+	while (list) {
+		ProcInfo *info = list->data;
+		GtkTreeIter child;
+		
+		if (!find_match_in_new_list (info->pid, pid_list, n))  {
+			removal_list  = g_list_append (removal_list, info);
+			info->queue = NEEDS_REMOVAL;
+			/* remove all children from tree */
+			if (info->visible) {
+				if (gtk_tree_model_iter_children (model, &child, &info->node))
+					remove_children_from_tree (procdata, model, &child);
+			}
+		}
 		list = g_list_next (list);
-		procdata->info = g_list_remove (procdata->info, oldinfo);
-		proctable_free_info (oldinfo);
 	}
 	
+	/* Add or remove processes from the tree */
+	list = procdata->info;
+	while (list) {
+		ProcInfo *info = list->data;
+		
+		if (info->queue == NEEDS_REMOVAL) {
+			remove_info_from_tree (info, procdata);
+			info->queue = NEEDS_NOTHING;
+		} else if (info->queue == NEEDS_ADDITION) {
+			insert_info_to_tree (info, procdata);
+			info->queue = NEEDS_NOTHING;
+		}
+		
+		list = g_list_next (list);
+	}
+	
+	/* Remove processes from the internal GList */
+	while(removal_list) {
+		ProcInfo *info = removal_list->data;
+		
+		remove_info_from_list (info, procdata);
+		
+		removal_list = g_list_next (removal_list);
+	}
+	
+	g_list_free (removal_list);	
 	
-
 }
 
 void
@@ -797,12 +842,10 @@ proctable_update_list (ProcData *data)
 	glibtop_cpu cpu;
 	gint which, arg;
 	gint n;
-	
-	
+
 	switch (procdata->config.whose_process) {
 	case ALL_PROCESSES:
-	case RUNNING_PROCESSES:
-	case FAVORITE_PROCESSES:
+	case ACTIVE_PROCESSES:
 		which = GLIBTOP_KERN_PROC_ALL;
 		arg = 0;
 		break;
@@ -812,7 +855,6 @@ proctable_update_list (ProcData *data)
 		break;
 	}
 	
-
 	pid_list = glibtop_get_proclist (&proclist, which, arg);
 	n = proclist.number;
 	
@@ -825,10 +867,8 @@ proctable_update_list (ProcData *data)
 	refresh_list (procdata, pid_list, n);
 	
 	g_free (pid_list);
-	
 }
 
-
 void
 proctable_update_all (ProcData *data)
 {
@@ -838,7 +878,6 @@ proctable_update_all (ProcData *data)
 	
 	if (procdata->config.show_more_info)
 		infoview_update (procdata);
-
 }
 
 void 
@@ -854,8 +893,6 @@ proctable_clear_tree (ProcData *data)
 	proctable_free_table (procdata);
 	
 	update_sensitivity (procdata, FALSE);
-	
-
 }
 
 void		
@@ -872,7 +909,6 @@ proctable_free_table (ProcData *procdata)
 	
 	g_list_free (procdata->info);
 	procdata->info = NULL;
-	
 }
 
 void
@@ -959,7 +995,5 @@ proctable_search_table (ProcData *procdata, gchar *string)
 	}
 	
 	increment --;
-
 }
-
 	
